'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
from webuntis import utils, objects, errors
from webuntis.utils import result_wrapper, log, rpc_request


class JSONRPCSession(object):
    '''Lower-level version of :py:class:`Session`. Do not use this.'''

    config = None
    '''Dictionary with configuration.'''

    def __init__(self, **kwargs):
        self.config = utils.FilterDict(utils.config_keys)
        config = {
            'server': None,
            'school': None,
            'useragent': None,
            'username': None,
            'password': None,
            'jsessionid': None,
            'login_repeat': 0
        }
        config.update(kwargs)
        self.config.update(config)

    def __enter__(self):
        '''Context-manager'''
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''Context-manager -- the only thing we need to clean up is to log out
        '''
        self.logout(suppress_errors=True)

    def logout(self, suppress_errors=False):
        '''
        Log out of session

        :type suppress_errors: bool
        :param suppress_errors: Whether to suppress errors.

        :raises: :py:class:`webuntis.errors.NotLoggedInError` -- Can't log out
            because not logged in. Raised unless ``suppress_errors`` is
            ``True``.
        '''
        def throw_errors():
            if not suppress_errors:
                raise errors.NotLoggedInError('We already were logged out.')

        try:
            # Send a JSON-RPC 'logout' method without parameters to log out
            self._request('logout')
        except errors.NotLoggedInError:
            throw_errors()

        try:
            del self.config['jsessionid']
        except KeyError:
            throw_errors()

    def login(self):
        '''Initializes an authentication, provided we have the credentials for
        it.

        :returns: The session. This is useful for jQuery-like command
            chaining::

                s = webuntis.Session(...).login()

        :raises: :py:class:`webuntis.errors.BadCredentialsError` --
            Username/Password missing or invalid.
        :raises: :py:class:`webuntis.errors.AuthError` -- Didn't recieve a
            session ID for unknown reasons.
        '''

        if 'username' not in self.config \
                or 'password' not in self.config:
            raise errors.BadCredentialsError('No login data specified.')

        log('debug', 'Trying to authenticate with username/password...')
        log('debug', 'Username: %s Password: %s' %
           (self.config['username'], self.config['password']))
        res = self._request('authenticate', {
            'user': self.config['username'],
            'password': self.config['password'],
            'client': self.config['useragent']
        }, use_login_repeat=False)
        log('debug', res)
        if 'sessionId' in res:
            log('debug', 'Did get a jsessionid from the server:')
            self.config['jsessionid'] = res['sessionId']
            log('debug', self.config['jsessionid'])
        else:
            raise errors.AuthError(
                'Something went wrong while authenticating',
                res
            )

        return self

    def _request(self, method, params=None, use_login_repeat=None):
        if use_login_repeat is None:
            use_login_repeat = (method not in ('logout', 'authenticate'))
        attempts_left = self.config['login_repeat'] if use_login_repeat else 0

        data = None

        while data is None:
            try:
                data = rpc_request(self.config, method, params)
            except errors.NotLoggedInError:
                if attempts_left > 0:
                    self.logout(suppress_errors=True)
                    self.login()
                else:
                    raise errors.NotLoggedInError(
                        'Tried to login several times, failed. Original method'
                        ' was ' + method)
            else:
                return data

            attempts_left -= 1  # new round!


class ResultWrapperMixin(object):
    @result_wrapper
    def departments(self):
        '''Get all departments.

        :rtype: :py:class:`webuntis.objects.DepartmentList`
        '''
        return objects.DepartmentList, 'getDepartments', {}

    @result_wrapper
    def holidays(self):
        '''Get all holidays.

        :rtype: :py:class:`webuntis.objects.HolidayList`
        '''
        return objects.HolidayList, 'getHolidays', {}

    @result_wrapper
    def klassen(self, schoolyear=None):
        '''Get all school classes.

        :param schoolyear: The schoolyear where the classes should be fetched
            from.
        :type schoolyear: :py:class:`webuntis.objects.SchoolyearObject` or an
            integer ID of it

        :rtype: :py:class:`webuntis.objects.KlassenList`
        '''
        params = {}
        if schoolyear:
            params['schoolyearId'] = int(schoolyear)

        return objects.KlassenList, 'getKlassen', params

    @result_wrapper
    def timetable(self, start, end, **type_and_id):
        '''Get the timetable for a specific school class and time period.

        :type start: :py:class:`datetime.datetime` or  :py:class:`datetime.date`
        :param start: The beginning of the time period.

        :type end: :py:class:`datetime.datetime` or  :py:class:`datetime.date`
        :param end: The end of the time period.

        :rtype: :py:class:`webuntis.objects.PeriodList`

        Furthermore you have to explicitly define a klasse, teacher, subject,
        room or student parameter containing the id or the object of the thing
        you want to get a timetable about::

            import datetime
            today = datetime.date.today()
            monday = today - datetime.timedelta(days=today.weekday())
            friday = monday + datetime.timedelta(days=4)

            klasse = s.klassen().filter(id=1)[0]  # schoolclass #1
            tt = s.timetable(klasse=klasse, start=monday, end=friday)

        :raises: :exc:`ValueError`, :exc:`TypeError`
        '''
        element_type_table = {
            'klasse':  1,
            'teacher': 2,
            'subject': 3,
            'room':    4,
            'student': 5
        }

        invalid_type_error = TypeError(
            'You have to specify exactly one of the following parameters by '
            'keyword: ' +
            (', '.join(element_type_table.keys()))
        )

        if len(type_and_id) != 1:
            raise invalid_type_error

        element_type, element_id = list(type_and_id.items())[0]

        element_type = utils.userinput.string(element_type)

        if element_type not in element_type_table:
            raise invalid_type_error

        # if we have to deal with an object in element_id,
        # its id gets placed here anyway

        json_start = utils.datetime_utils.format_date(start)
        json_end =   utils.datetime_utils.format_date(end)

        if json_start > json_end:
            raise ValueError('Start can\'t be later than the end.')

        parameters = {
            'id': int(element_id),
            'type': element_type_table[element_type],
            'startDate': json_start,
            'endDate': json_end
        }

        return objects.PeriodList, 'getTimetable', parameters

    @result_wrapper
    def rooms(self):
        '''Get all rooms of a school.

        :rtype: :py:class:`webuntis.objects.RoomList`
        '''
        return objects.RoomList, 'getRooms', {}

    @result_wrapper
    def schoolyears(self):
        '''Get all schoolyears.

        :rtype: :py:class:`webuntis.objects.SchoolyearList`
        '''
        return objects.SchoolyearList, 'getSchoolyears', {}

    @result_wrapper
    def subjects(self):
        '''Get all subjects.

        :rtype: :py:class:`webuntis.objects.SubjectList`
        '''
        return objects.SubjectList, 'getSubjects', {}

    @result_wrapper
    def teachers(self):
        '''Get all teachers.

        :rtype: :py:class:`webuntis.objects.TeacherList`
        '''
        return objects.TeacherList, 'getTeachers', {}

    @result_wrapper
    def timegrid(self):
        '''Get a "timegrid", whatever the hell that is. According to the
        official API docs, it is supposed to be useful when generating your own
        timetable. Maybe :py:meth:`webuntis.objects.PeriodList.to_table` could
        make use of this one day.

        :rtype: :py:class:`webuntis.objects.TimeunitList`
        '''
        return objects.TimeunitList, 'getTimegridUnits', {}

    @result_wrapper
    def statusdata(self):
        '''Information about lesson types and period codes, specifically about
        the colors used to highlight them in the web-interface of WebUntis.

        :rtype: :py:class:`webuntis.objects.StatusData`
        '''
        return objects.StatusData, 'getStatusData', {}


class Session(JSONRPCSession, ResultWrapperMixin):
    '''The origin of everything you want to do with the WebUntis API. Can be
    used as a context-handler.

    Configuration can be set with keyword arguments when initializing
    :py:class:`Session`. Unless noted otherwise, they get saved in a dictionary
    located in the instance's :py:attr:`config` attribute and can be modified
    afterwards.

    :type username: str
    :param username: The username used for the API.

    :type password: str
    :param password: The password used for the API.

    :type server: str
    :param server: A host name, a URL, or a URL without path.

            >>> s = webuntis.Session(..., server='thalia.webuntis.com')
            >>> s.config['server']
            'http://thalia.webuntis.com/WebUntis/jsonrpc.do'
            >>> # notice that there's NO SLASH at the end!
            >>> s.config['server'] = 'https://thalia.webuntis.com'
            >>> s.config['server']
            'https://thalia.webuntis.com/WebUntis/jsonrpc.do'
            >>> s.config['server'] = 'https://thalia.webuntis.com/'
            >>> # because a slash gets interpreted as the full path to the API
            >>> # endpoint, which would crash during login
            >>> s.config['server']
            'http://thalia.webuntis.com/'
            >>> s.config['server'] = '!"$%/WebUntis/jsonrpc.do'
            Traceback blah blah something ValueError

    :type school: str
    :param school: A valid school name.

    :type useragent: str
    :param useragent: A string containing a useragent. Please include useful
        information, such as an email address, for the server maintainer. Just
        like you would do with the HTTP useragents of bots.

    :type cachelen: int
    :param cachelen: The maximum size of the internal cache. All results are
        saved in it, but they only get used if you set the ``from_cache``
        parameter on a session method to ``True``. This parameter is not saved
        in the configuration dictionary.

        ::

            s.timetable(klasse=123)  # saves in cache
            s.timetable(klasse=123)  # fetch data again, override old value
            s.timetable(klasse=123, from_cache=True)  # get directly from cache

        The reason this cache was added is that the API only allows you to
        fetch a whole list of objects (teachers/schoolclasses/...), not single
        ones. It would seriously harm performance to fetch the whole list each
        time we want information about a single object. Without the cache, i
        sometimes experienced a performance decrease about twenty seconds, so i
        wouldn't set the ``cachelen`` to anything smaller than ``5``.

        Default value is ``20``.

        You can clear the cache using::

            s.cache.clear('timetable')  # clears all cached timetables
            s.cache.clear()  # clears everything from the cache

    :type jsessionid: str
    :param jsessionid: The session key to use. You usually shouldn't touch
        this.

    :type login_repeat: int
    :param login_repeat: The amount of times `python-webuntis` should try to
        login when finding no or an expired session. Default to ``0``, meaning
        it won't do that.

    '''

    cache = None
    '''Contains the caching dictionary for requests.'''

    # Repeated here because sphinx doesn't recognize it when defined in
    # JSONRPCSession:
    config = None
    '''The config dictionary, filled with most keyword arguments from
    initialization.'''

    def __init__(self, **config):
        try:
            cachelen = config['cachelen']
            del config['cachelen']
        except KeyError:
            cachelen = 20

        self.cache = utils.SessionCache(maxlen=cachelen)

        JSONRPCSession.__init__(self, **config)



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
from datetime import datetime


def table(periods, dates=None, times=None):
    '''The backend of :py:meth:`webuntis.objects.PeriodList.to_table`.'''

    if not len(periods):
        return []

    # generate some useful sets
    times = set(times or []).union(period.start.time() for period in periods)
    dates = set(dates or []).union(period.start.date() for period in periods)
    datetimes = set(datetime.combine(d, t) for d in dates for t in times)

    # create an empty table from all possible combinations of dates and times
    table = dict((t, dict((d, set()) for d in dates)) for t in times)

    # add the periods to the table
    # periods may be added twice if they are longer than one hour
    for period in periods:
        for dt in datetimes:
            if period.start <= dt and period.end > dt:
                table[dt.time()][dt.date()].add(period)

    # Convert the hashtable to the output format by sorting each dictionary's
    # .items() by key.
    return sorted((time, sorted(row.items())) for time, row in table.items())



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''


from __future__ import unicode_literals

import unittest
import mock
import os
import json
import webuntis
import logging
from copy import deepcopy

try:
    from StringIO import StringIO  # Python 2
except ImportError:
    from io import BytesIO as StringIO  # Python 3

tests_path = os.path.abspath(os.path.dirname(__file__))
data_path = tests_path + '/static'


def get_json_resource(name):
        with open(os.path.join(data_path, name)) as f:
            return json.load(f)


class TestCaseBase(unittest.TestCase):
    pass

stub_session_parameters = {
    'useragent': 'fooagent',
    'school': 'fooschool',
    'username': 'foouser',
    'password': 'hunter2',
    'server': 'fooboo_server',
    'jsessionid': 'FOOBOO_SESSION'
}


class OfflineTestCase(TestCaseBase):
    def setUp(self):
        def cb(*args, **kwargs):  # pragma: no cover
            raise Exception('These are offline tests.')

        self.request_patcher = patcher = \
                mock.patch('webuntis.utils.remote._send_request', new=cb)
        patcher.start()

        self.session = webuntis.Session(**stub_session_parameters)

    def tearDown(self):
        try:
            self.request_patcher.stop()
        except RuntimeError:
            logging.warning(
                'Failed to tear the request_patcher down properly.')

        self.session = None


def mock_results(methods, swallow_not_found=False):
    '''Mock API methods more easily.

    :type methods: dict
    :param methods: A dictionary containing one callable for each API method.

    :type swallow_not_found: bool
    :param swallow_not_found: Whether to return {'result': {}} on unmocked API
        methods.
    '''
    def new(url, jsondata, headers):
        method = jsondata['method']
        try:
            method_mock = methods[method]
        except KeyError:  # pragma: no cover
            if not swallow_not_found:
                raise
            else:
                data = {'result': {}}
        else:
            if not hasattr(method_mock, 'calls'):
                method_mock.calls = []
            method_mock.calls.append((url, jsondata, headers))
            data = method_mock(url, jsondata, headers)

        d = {'id': jsondata['id']}
        d.update(data)

        return deepcopy(d)

    return mock.patch('webuntis.utils.remote._send_request', new=new)


def raw_vs_object(jsondata, result):
    '''zip json data and results, but grouped by id instead of order. Also runs
    some checks that hashes are unique.'''
    raw_lookup = dict((x['id'], x) for x in jsondata)
    known_hashes = set()

    for obj in result:
        assert hash(obj) not in known_hashes, obj.id
        known_hashes.add(hash(obj))

        yield (raw_lookup[obj.id], obj)

    assert len(known_hashes) == len(jsondata) == len(result), \
        (len(known_hashes), len(jsondata), len(result))


def mock_urlopen(data, expected_url, expected_data, expected_headers):
    def mocking_func(requestobj):
        given_url = requestobj.get_full_url()
        given_data = requestobj.data
        given_headers = dict(requestobj.header_items())

        assert given_url == expected_url, given_url
        assert given_data == expected_data, given_data
        assert given_headers == expected_headers, given_headers

        if isinstance(data, Exception):
            raise data

        io = StringIO(data)
        return io

    return mock.patch(
        'webuntis.utils.third_party.urlrequest.urlopen',
        new=mocking_func
    )



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

_logger = None


def log(level, message, *args, **kwargs):
    '''Log a message to the logger used. Written inside a function so it can be
    overridden if really neccessary.'''
    global _logger
    if _logger is None:
        import logging
        _logger = logging.getLogger('webuntis')

    getattr(_logger, level)(message, *args, **kwargs)



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from .userinput import config_keys
from .misc import FilterDict, \
    SessionCache, \
    LruDict, \
    cache_key, \
    lazyproperty, \
    result_wrapper
from .logger import log
from .remote import rpc_request



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
from webuntis.utils import datetime_utils, lazyproperty, \
    timetable_utils


class Result(object):
    '''Base class used to represent most API objects.

    :param data: Usually JSON data that should be represented.

                 In the case of :py:class:`ListResult`, however, it might also
                 be a list of JSON mixed with :py:class:`ListItem` objects.

    :param parent: (optional) A result object this result should be the child
                    of. If given, the session will be inherited.

    :param session: Mandatory if ``parent`` is not supplied. Overrides the
                    parent's inherited session.

    '''

    _parent = None
    _session = None
    _data = None

    def __init__(self, data, parent=None, session=None):
        if not isinstance(parent, Result) and parent is not None:
            raise TypeError('If provided, parent must be an instance of '
                            'webuntis.objects.Result.')

        if bool(parent is None) == bool(session is None):
            raise TypeError('Either parent or session has to be provided.')

        self._session = session or parent._session
        self._parent = parent
        self._data = data

    @lazyproperty
    def id(self):
        '''The ID of this element.

        An ID is needed for the object to be hashable. Therefore a result
        may bring its own implementation of this method even though the
        original API response didn't contain any ID.'''
        return self._data['id'] if 'id' in self._data else None

    def __int__(self):
        '''This is useful if the users pass a ListItem when a numerical ID
        is expected, so we just can put the thing through int(), regardless of
        what type it is.'''
        return self.id

    def __hash__(self):
        return hash('%s#%i' % (self.__class__.__name__, self.id))


class ListItem(Result):
    '''ListItems represent an item in a
    :py:class:`Result`. They don\'t contain methods to
    retrieve data.'''


class ListResult(Result):
    '''A list-like version of :py:class:`Result` that takes a list and returns
    a list of objects, containing a list value each.

        '''

    # When the Result returns an array, this is very useful. Every item of that
    # array will be fed to an instance of self._itemclass, with the session and
    # the array item as initialization arguments.

    #: the class which should be used to instantiate an array item.
    _itemclass = ListItem

    def filter(self, **criterions):
        '''
        Return a list of all objects, filtered by attributes::

            foo = s.klassen().filter(id=1)  # is kind-of the same as
            foo = [kl for kl in s.klassen() if kl.id == 1]

            # We can also use sets to match multiple values.
            bar = s.klassen().filter(name={'1A', '2A', '3A', '4A'})
            # is kind-of the same as
            bar = [kl for kl in s.klassen()
                   if kl.id in {'1A', '2A', '3A', '4A'}]

            # Since ``filter`` returns a ListResult itself too, we can chain
            # multiple calls together:
            bar = s.klassen().filter(id=4, name='7A')  # is the same as
            bar = s.klassen().filter(id=4).filter(name='7A')

        :py:meth:`filter` is also used when using the ``in`` operator on a
        :py:class:`ListResult`::

            we_have_it = {'name': '6A'} in s.klassen()  # same as
            we_have_it = bool(s.klassen().filter(name='6A'))


        .. note::
            This is only available because it looks nicer than list
            comprehensions or generator expressions. Depending on your usecase
            alternatives to this method may be faster.

        '''
        criterions = list(criterions.items())

        def meets_criterions(item):
            '''Returns true if the item meets the criterions'''
            for key, value in criterions:
                # if the attribute value isn't one we're looking for
                attribute = getattr(item, key)
                if attribute == value:
                    continue
                elif isinstance(value, set) and attribute in value:
                    continue
                else:
                    return False

            return True

        return type(self)(
            parent=self,
            data=[item for item in self if meets_criterions(item)]
        )

    def __contains__(self, criterion):
        return bool(self.filter(**criterion))

    def __getitem__(self, i):
        '''Makes the object iterable and behave like a list'''
        data = self._data[i]  # fails if there is no such item

        if type(data) is not self._itemclass:
            data = self._data[i] = self._itemclass(
                parent=self,
                data=data
            )

        return data

    def __len__(self):
        '''Return the length of the items'''
        return len(self._data)


class DepartmentObject(ListItem):
    '''Represents a department'''

    @lazyproperty
    def name(self):
        '''short name such as *R1A*'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name, such as *Raum Erste A*. Not predictable.'''
        return self._data['longName']


class DepartmentList(ListResult):
    '''A list of departments, in form of :py:class:`DepartmentObject`
    instances.'''
    _itemclass = DepartmentObject


class HolidayObject(ListItem):
    '''Represents a single holiday.'''

    @lazyproperty
    def start(self):
        '''The start date of the holiday, as a datetime object.'''
        return datetime_utils.parse_date(self._data['startDate'])

    @lazyproperty
    def end(self):
        '''The end of the holiday'''
        return datetime_utils.parse_date(self._data['endDate'])

    @lazyproperty
    def name(self):
        '''Name, such as *Nationalfeiertag*.'''
        return self._data['longName']

    @lazyproperty
    def short_name(self):
        '''Abbreviated form of the name'''
        return self._data['name']


class HolidayList(ListResult):
    '''A list of holidays, in form of :py:class:`HolidayObject`
    instances.'''
    _itemclass = HolidayObject


class KlassenObject(ListItem):
    '''Represents a school class.'''

    @lazyproperty
    def name(self):
        '''Name of class'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name of class'''
        return self._data['longName']


class KlassenList(ListResult):
    '''A list of school classes, in form of :py:class:`KlassenObject`
    instances.'''
    _itemclass = KlassenObject


class PeriodObject(ListItem):
    '''Represents a time range, where lessons/subjects may be held.'''

    @lazyproperty
    def start(self):
        '''The start date/time of the period, as datetime object.'''

        return datetime_utils.parse_datetime(
            self._data['date'],
            self._data['startTime']
        )

    @lazyproperty
    def end(self):
        '''The end date/time of the period.'''

        return datetime_utils.parse_datetime(
            self._data['date'],
            self._data['endTime']
        )

    @lazyproperty
    def klassen(self):
        '''A :py:class:`KlassenList` containing the classes which are attending
        this period.'''

        return self._session.klassen(from_cache=True).filter(
            id=set([kl['id'] for kl in self._data['kl']])
        )

    @lazyproperty
    def teachers(self):
        '''A list of :py:class:`TeacherObject` instances,
        which are attending this period.'''

        return self._session.teachers(from_cache=True).filter(
            id=set([te['id'] for te in self._data['te']])
        )

    @lazyproperty
    def subjects(self):
        '''A :py:class:`SubjectList` containing the subjects which are topic of
        this period. This is not used for things like multiple language lessons
        (*e.g.* Latin, Spanish, French) -- each of those will get placed in
        their own period.'''

        return self._session.subjects(from_cache=True).filter(
            id=set([su['id'] for su in self._data['su']])
        )

    @lazyproperty
    def rooms(self):
        '''The rooms (:py:class:`RoomList`) where this period is taking place
        at. This also is not used for multiple lessons, but rather for a single
        lesson that is actually occuring at multiple locations (?).'''

        return self._session.rooms(from_cache=True).filter(
            id=set([ro['id'] for ro in self._data['ro']])
        )

    @lazyproperty
    def code(self):
        '''May be:

          - ``None`` -- There's nothing special about this period.
          - ``"cancelled"`` -- Cancelled
          - ``"irregular"`` -- Substitution/"Supplierung"/Not planned event
        '''

        if 'code' not in self._data:
            return None
        else:
            return self._data['code'] or None

    @lazyproperty
    def type(self):
        '''May be:

          - ``"ls"`` -- Normal lesson
          - ``"oh"`` -- Office hour
          - ``"sb"`` -- Standby
          - ``"bs"`` -- Break Supervision
          - ``"ex"`` -- Examination
        '''

        return (self._data['lstype'] if 'lstype' in self._data else 'ls')


class PeriodList(ListResult):
    '''Aka timetable, a list of periods, in form of :py:class:`PeriodObject`
    instances.'''
    _itemclass = PeriodObject

    def to_table(self, dates=None, times=None):
        '''
        Creates a table-like structure out of the periods. Useful for rendering
        timetables in HTML and other markup languages.

        Check out the example from the repository for clarification.

        :param dates: An iterable of :py:class:`datetime.date` objects that
            definetly should be included in the table. If this parameter is
            ``None``, the timetable is just as wide as it has to be, leaving
            out days without periods.

        :param times: An iterable of :py:class:`datetime.time` objects that
            definetly should be included in the table. If this parameter is
            ``None``, the timetable is just as tall as it has to be, leaving
            out hours without periods.

        :returns: A list containing "rows", which in turn contain "hours",
            which contain :py:class:`webuntis.objects.PeriodObject` instances
            which are happening at the same time.

        '''

        return timetable_utils.table(self, dates=dates, times=times)


class RoomObject(ListItem):
    '''Represents a physical room. Such as a classroom, but also the physics
    lab or whatever.
    '''
    @lazyproperty
    def name(self):
        '''The short name of the room. Such as PHY.'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''The long name of the room. Such as "Physics lab".'''
        return self._data['longName']


class RoomList(ListResult):
    '''A list of rooms, in form of :py:class:`RoomObject` instances.'''
    _itemclass = RoomObject


class SchoolyearObject(ListItem):
    '''Represents a schoolyear.'''

    @lazyproperty
    def name(self):
        '''"2010/2011"'''

        return self._data['name']

    @lazyproperty
    def start(self):
        '''The start date of the schoolyear, as datetime object'''
        return datetime_utils.parse_date(self._data['startDate'])

    @lazyproperty
    def end(self):
        '''The end date'''
        return datetime_utils.parse_date(self._data['endDate'])

    @lazyproperty
    def is_current(self):
        '''
        Boolean, check if this is the current schoolyear::

            >>> y = s.schoolyears()
            >>> y.current.id
            7
            >>> y.current.is_current
            True
            >>> y.filter(id=y.current.id).is_current
            True

        '''
        return (self == self._parent.current)


class SchoolyearList(ListResult):
    '''A list of schoolyears, in form of :py:class:`SchoolyearObject`
    instances.'''
    _itemclass = SchoolyearObject

    @lazyproperty
    def current(self):
        '''Returns the current schoolyear in form of a
        :py:class:`SchoolyearObject`'''
        current_data = self._session._request('getCurrentSchoolyear')
        current = self.filter(id=current_data['id'])[0]
        return current


class SubjectObject(ListItem):
    '''Represents a subject.'''

    @lazyproperty
    def name(self):
        '''Short name of subject, such as *PHY*'''
        return self._data['name']

    @lazyproperty
    def long_name(self):
        '''Long name of subject, such as *Physics*'''
        return self._data['longName']


class SubjectList(ListResult):
    '''A list of subjects, in form of :py:class:`SubjectObject` instances.'''
    _itemclass = SubjectObject


class TeacherObject(ListItem):
    '''Represents a teacher.'''
    @lazyproperty
    def fore_name(self):
        '''fore name of the teacher'''
        return self._data['foreName']

    @lazyproperty
    def long_name(self):
        '''surname of teacher'''
        return self._data['longName']

    surname = long_name

    @lazyproperty
    def name(self):
        '''full name of the teacher'''
        return self._data['name']


class TeacherList(ListResult):
    '''A list of teachers, in form of :py:class:`TeacherObject` instances.'''
    _itemclass = TeacherObject


class TimeunitObject(ListItem):
    '''A bunch of timeunits for a specific day.'''

    @lazyproperty
    def times(self):
        '''A list of tuples containing the start and the end of each timeunit
        as datetime '''

        return [
            (
                datetime_utils.parse_time(unit['startTime']),
                datetime_utils.parse_time(unit['endTime'])
            ) for unit in self._data['timeUnits']
        ]

    @lazyproperty
    def day(self):
        '''The day the timeunit list is for'''
        return self._data['day']

    @lazyproperty
    def id(self):
        return int(self.day)


class TimeunitList(ListResult):
    '''A list of times and dates for the current week, in form of
    :py:class:`TimeunitObject` instances. Doesn't contain actual data, but is
    useful when you want to generate a timetable::

        >>> grid = s.timegrid()
        >>>
        >>> # 1 = Sunday
        >>> # 2 = Monday
        >>> # ...
        >>> # 7 = Saturday
        >>> grid[0].day
        2
        >>> grid[0].times
        [
            (datetime.datetime(...), datetime.datetime(...)),
            ...
        ]

    .. note::
        The date-specific properties of the datetime objects are invalid, since
        these are not provided by the official API.
    '''

    _itemclass = TimeunitObject


class ColorInfo(Result):
    '''
    An object containing information about a lesson type or a period code::

        >>> lstype = s.statusdata().lesson_types[0]
        >>> lstype.name
        'ls'
        >>> lstype.forecolor
        '000000'
        >>> lstype.backcolor
        'ee7f00'

    ::

        >>> pcode = s.statusdata().period_codes[0]
        >>> pcode.name
        'cancelled'
        >>> pcode.forecolor
        'FFFFFF'
        >>> pcode.backcolor
        'FF0000'

    '''

    @lazyproperty
    def id(self):
        return hash(self.__class__.__name__ + self.name)

    @lazyproperty
    def name(self):
        '''The name of the LessonType or PeriodCode'''
        return list(self._data.items())[0][0]

    @lazyproperty
    def forecolor(self):
        '''The foreground color used in the web interface and elsewhere'''
        return self._data[self.name]['foreColor']

    @lazyproperty
    def backcolor(self):
        '''The background color used in the web interface and elsewhere'''
        return self._data[self.name]['backColor']


class StatusData(Result):
    '''Information about lesson types and period codes and their colors.'''

    @lazyproperty
    def lesson_types(self):
        '''A list of :py:class:`ColorInfo` objects, containing
        information about all lesson types defined'''
        return [
            ColorInfo(parent=self, data=data)
            for data in self._data['lstypes']
        ]

    @lazyproperty
    def period_codes(self):
        '''A list of :py:class:`ColorInfo` objects, containing
        information about all period codes defined'''
        return [
            ColorInfo(parent=self, data=data)
            for data in self._data['codes']
        ]



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
__version__ = '0.1.7'
from webuntis.session import Session



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
# Uncategorized utils go here

from __future__ import unicode_literals
from functools import wraps
from copy import deepcopy
from .third_party import OrderedDict


class lazyproperty(object):
    '''A read-only @property that is only evaluated once. Only usable on class
    instances' methods.

    Stolen from http://www.reddit.com/r/Python/comments/ejp25/cached_property_decorator_that_is_memory_friendly/
    '''
    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:  # pragma: no cover
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result


class LruDict(OrderedDict):
    def __init__(self, maxlen=50):
        super(LruDict, self).__init__()
        self._maxlen = maxlen

    def __setitem__(self, key, value):
        super(LruDict, self).__setitem__(key, value)
        while len(self.items()) > self._maxlen:
            self.popitem(last=False)

class SessionCache(LruDict):
    def clear(self, method=None):
        if method is None:
            LruDict.clear(self)
        else:
            for k in list(self):
                if k[0] == method:
                    del self[k]


class FilterDict(object):
    '''A dictionary which passes new values to a function found at the
    corresponding key in self.filters

    :param filters: A dictionary containing functions. If a new key is set into
    an instance of FilterDict, the filter dictionary will be consulted to
    filter the value with a function.

    >>> config = FilterDict({
    ...    'foo': lambda x: 'whoopdeedoo'
    ... })
    >>>
    >>> config['foo'] = 'somethingelse'
    >>> config['foo']
    'whoopdeedoo'

    '''
    filters = None
    _contents = None

    def __init__(self, filters):
        self.filters = filters
        self._contents = {}

    def __getitem__(self, name):
        # check if we got a real Option subclass
        if name in self._contents and self._contents[name] is not None:
            # every Option subclass has this
            return self._contents[name]
        elif name in self.filters:
            raise KeyError('No value for key: ' + name)
        else:
            raise KeyError('No value or filter for key: ' + name)

    def __setitem__(self, key, value):
        if value is None:
            if key in self._contents:
                del self._contents[key]
            return

        new_value = self.filters[key](value)

        if new_value is None:
            if key in self._contents:
                del self._contents[key]
            return

        self._contents[key] = new_value

    def __delitem__(self, key):
        del self._contents[key]

    def update(self, new_pairs):
        for key, value in new_pairs.items():
            self[key] = value

    def __contains__(self, key):
        return key in self._contents

    def __iter__(self):
        return iter(self._contents)


def result_wrapper(func):
    '''A decorator for the session methods that return result objects. The
    decorated function has to return a tuple with the result class to
    instantiate, a JSON-RPC method and its parameters.  This decorator fetches
    the data with the JSON-RPC method and parameters and returns an instance of
    the result class the inner function returned (and saves it in the session
    cache).
    '''
    @wraps(func)
    def inner(self, **kwargs):
        from_cache = False
        if 'from_cache' in kwargs and kwargs['from_cache']:
            from_cache = True
            del kwargs['from_cache']

        result_class, jsonrpc_method, jsonrpc_args = func(self, **kwargs)
        key = cache_key(func.__name__, jsonrpc_args)

        if from_cache and key in self.cache:
            return self.cache[key]

        data = self._request(jsonrpc_method, jsonrpc_args)
        self.cache[key] = result = result_class(session=self, data=data)
        return result

    return inner


def cache_key(method, args=None):
    '''Get a hashable object given a string and a dictionary.'''
    if args is None:
        args = {}
    hash_args = frozenset(deepcopy(args).items())

    return (method, hash_args)



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
import mock
import webuntis
import datetime

from webuntis.tests._utils import OfflineTestCase, mock_results, \
    get_json_resource, stub_session_parameters, raw_vs_object, mock_urlopen

from nose.tools import eq_, assert_not_equal, assert_raises


class DataFetchingTests(OfflineTestCase):
    '''Test all Result objects from a high level.'''

    def test_getdepartments(self):
        jsonstr = get_json_resource('getdepartments_mock.json')

        def getDepartments(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getDepartments': getDepartments}

        with mock_results(methods):
            for dep_raw, dep in raw_vs_object(
                    jsonstr, self.session.departments()):
                eq_(dep_raw['id'], dep.id)
                eq_(dep_raw['longName'], dep.long_name)
                eq_(dep_raw['name'], dep.name)

    def test_getholidays(self):
        jsonstr = get_json_resource('getholidays_mock.json')

        def getHolidays(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getHolidays': getHolidays}

        with mock_results(methods):
            for holiday_raw, holiday in raw_vs_object(
                    jsonstr, self.session.holidays()):
                eq_(holiday_raw['id'], holiday.id)
                eq_(holiday_raw['name'], holiday.short_name)
                eq_(holiday_raw['longName'], holiday.name)

                eq_(holiday_raw['startDate'],
                    int(holiday.start.strftime('%Y%m%d')))
                eq_(holiday_raw['endDate'],
                    int(holiday.end.strftime('%Y%m%d')))

    def test_getklassen(self):
        jsonstr = get_json_resource('getklassen_mock.json')
        schoolyear_id = 123

        def getKlassen(url, jsondata, headers):
            if len(methods['getKlassen'].calls) == 2:
                eq_(jsondata['params']['schoolyearId'], schoolyear_id)

            assert len(methods['getKlassen'].calls) <= 2
            return {'result': jsonstr}

        methods = {'getKlassen': getKlassen}

        with mock_results(methods):
            klassen = self.session.klassen()
            for klasse_raw, klasse in raw_vs_object(jsonstr, klassen):
                eq_(klasse_raw['id'], klasse.id)
                eq_(klasse_raw['name'], klasse.name)
                eq_(klasse_raw['longName'], klasse.long_name)

            eq_(klassen.filter(id=129)[0].id, 129)
            eq_([129, 130, 137], sorted(
                kl.id for kl in
                klassen.filter(id=set([129, 130, 137]))
                ))

            assert {'id': 129} in klassen
            assert {'name': '6A'} in klassen

            # calling a second time for validating the schoolyear
            self.session.klassen(schoolyear=schoolyear_id)

    def test_gettimetables(self):
        jsonstr = get_json_resource('gettimetables_mock.json')
        jsonstr_kl = get_json_resource('getklassen_mock.json')
        jsonstr_te = get_json_resource('getteachers_mock.json')
        jsonstr_su = get_json_resource('getsubjects_mock.json')
        jsonstr_ro = get_json_resource('getrooms_mock.json')

        def getTimetable(url, jsondata, headers):
            return {'result': jsonstr}

        def getKlassen(url, jsondata, headers):
            return {'result': jsonstr_kl}

        def getTeachers(url, jsondata, headers):
            return {'result': jsonstr_te}

        def getSubjects(url, jsondata, headers):
            return {'result': jsonstr_su}

        def getRooms(url, jsondata, headers):
            return {'result': jsonstr_ro}

        methods = {
            'getTimetable': getTimetable,
            'getKlassen': getKlassen,
            'getTeachers': getTeachers,
            'getSubjects': getSubjects,
            'getRooms': getRooms
        }

        with mock_results(methods):
            tt = self.session.timetable(klasse=114,
                                        start='20120301', end='20120301')

            for period_raw, period in raw_vs_object(jsonstr, tt):
                eq_(int(period.start.strftime('%H%M')),
                    period_raw['startTime'])
                eq_(int(period.start.strftime('%Y%m%d')),
                    period_raw['date'])
                eq_(int(period.end.strftime('%H%M')),
                    period_raw['endTime'])
                eq_(int(period.end.strftime('%H%M')),
                    period_raw['endTime'])

                if 'code' in period_raw:
                    eq_(period_raw['code'], period.code)
                else:
                    eq_(period.code, None)

                if 'lstype' in period_raw:
                    eq_(period_raw['lstype'], period.type)
                else:
                    eq_(period.type, 'ls')

                eq_(len(period_raw['kl']), len(period.klassen))
                for klasse_raw, klasse in raw_vs_object(
                        period_raw['kl'], period.klassen):
                    eq_(klasse.id, klasse_raw['id'])

                eq_(len(period_raw['te']), len(period.teachers))
                for teacher_raw, teacher in raw_vs_object(
                        period_raw['te'], period.teachers):
                    eq_(teacher.id, teacher_raw['id'])

                eq_(len(period_raw['su']), len(period.subjects))
                for subject_raw, subject in raw_vs_object(
                        period_raw['su'], period.subjects):
                    eq_(subject.id, subject_raw['id'])

                eq_(len(period_raw['ro']), len(period.rooms))
                for room_raw, room in raw_vs_object(
                        period_raw['ro'], period.rooms):
                    eq_(room.id, room_raw['id'])

            def validate_table(periods):
                counter = 0
                table = periods.to_table()
                for time, row in table:
                    for date, cell in row:
                        for hour in cell:
                            counter += 1

                eq_(counter, len(periods))

            validate_table(tt)
            eq_(len(webuntis.utils.timetable_utils.table([])), 0)

    def test_gettimetables_invalid_type_and_id(self):
        assert_raises(TypeError, self.session.timetable,
                      start=None, end=None, klasse=123, teacher=123)
        assert_raises(TypeError, self.session.timetable,
                      start=None, end=None)
        assert_raises(TypeError, self.session.timetable,
                      start=None, end=None, foo=123)

    def test_gettimetables_start_later_than_end(self):
        start = datetime.datetime.strptime('20130101', '%Y%m%d')
        end = datetime.datetime.strptime('20120101', '%Y%m%d')

        assert_raises(ValueError, self.session.timetable,
                      start=start, end=end, klasse=123)

    def test_gettimetables_start_and_end_are_none(self):
        assert_raises(TypeError, self.session.timetable,
                      start=None, end=None, klasse=123)

    def test_getrooms(self):
        jsonstr = get_json_resource('getrooms_mock.json')

        def getRooms(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getRooms': getRooms}

        with mock_results(methods):
            for room_raw, room in raw_vs_object(jsonstr, self.session.rooms()):
                eq_(room_raw['longName'], room.long_name)
                eq_(room_raw['name'], room.name)
                eq_(room_raw['id'], room.id)

    def test_getschoolyears(self):
        jsonstr = get_json_resource('getschoolyears_mock.json')
        current_json = jsonstr[3]

        def getSchoolyears(url, jsondata, headers):
            return {'result': jsonstr}

        def getCurrentSchoolyear(url, jsondata, headers):
            return {'result': current_json}

        methods = {
            'getSchoolyears': getSchoolyears,
            'getCurrentSchoolyear': getCurrentSchoolyear
        }

        with mock_results(methods):
            schoolyears = self.session.schoolyears()

            eq_(current_json['id'], schoolyears.current.id)
            assert schoolyears.current.is_current
            current_count = 0
            for year_raw, year in raw_vs_object(jsonstr, schoolyears):
                eq_(year_raw['id'], year.id)
                eq_(year_raw['name'], year.name)

                eq_(
                    year_raw['startDate'],
                    int(year.start.strftime('%Y%m%d'))
                )
                eq_(
                    year_raw['endDate'],
                    int(year.end.strftime('%Y%m%d'))
                )
                if year.is_current:
                    eq_(year, schoolyears.current)
                    current_count += 1
                    eq_(current_count, 1)

    def test_getsubjects(self):
        jsonstr = get_json_resource('getsubjects_mock.json')

        def getSubjects(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getSubjects': getSubjects}

        with mock_results(methods):
            for subj_raw, subj in raw_vs_object(
                    jsonstr, self.session.subjects()):
                eq_(subj_raw['id'], subj.id)
                eq_(subj_raw['name'], subj.name)
                eq_(subj_raw['longName'], subj.long_name)

    def test_getteachers(self):
        jsonstr = get_json_resource('getteachers_mock.json')

        def getTeachers(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getTeachers': getTeachers}

        with mock_results(methods):
            for t_raw, t in raw_vs_object(jsonstr, self.session.teachers()):
                eq_(t_raw['longName'], t.long_name)
                eq_(t_raw['longName'], t.surname)
                eq_(t_raw['foreName'], t.fore_name)
                eq_(t_raw['name'], t.name)

    def test_gettimegrid(self):
        jsonstr = get_json_resource('gettimegrid_mock.json')

        def getTimegridUnits(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getTimegridUnits': getTimegridUnits}

        with mock_results(methods):
            for t_raw, t in zip(jsonstr, self.session.timegrid()):
                assert t_raw['day'] == t.day == t.id
                for t2_raw, t2 in zip(t_raw['timeUnits'], t.times):
                    eq_(t2_raw['startTime'], int(t2[0].strftime('%H%M')))
                    eq_(t2_raw['endTime'], int(t2[1].strftime('%H%M')))

    def test_getstatusdata(self):
        jsonstr = get_json_resource('getstatusdata_mock.json')

        def getStatusData(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getStatusData': getStatusData}

        def validate_statusdata(raw, processed):
            name = list(raw.items())[0][0]
            colors = raw[name]
            eq_(name, processed.name)
            eq_(colors['foreColor'], processed.forecolor)
            eq_(colors['backColor'], processed.backcolor)
            assert type(processed.id) is int

        with mock_results(methods):
            statusdata = self.session.statusdata()
            for lstype_raw, lstype in zip(
                    jsonstr['lstypes'], statusdata.lesson_types):
                validate_statusdata(lstype_raw, lstype)

            for code_raw, code in zip(
                    jsonstr['codes'], statusdata.period_codes):
                validate_statusdata(code_raw, code)


class SessionUsageTests(OfflineTestCase):
    '''Test features of webuntis.Session.'''
    def test_login_repeat_invalid_session(self):
        retry_amount = 5
        calls = []

        # This produces a list of 5 * ['getCurrentSchoolyear'] with ['logout',
        # 'authenticate'] between each.
        expected_calls = (
            ['getCurrentSchoolyear', 'logout', 'authenticate']
            * (retry_amount + 1)
        )[:-2]

        def authenticate(url, jsondata, headers):
            calls.append(jsondata['method'])
            return {
                'result': {'sessionId': 'Foobar_session_' + jsondata['id']}
            }

        def getCurrentSchoolyear(url, jsondata, headers):
            calls.append(jsondata['method'])
            return {
                'error': {'code': -8520, 'message': 'Not Logged In!'}
            }

        def logout(url, jsondata, headers):
            calls.append(jsondata['method'])
            return {
                'result': {'bla': 'blub'}
            }

        methods = {
            'authenticate': authenticate,
            'getCurrentSchoolyear': getCurrentSchoolyear,
            'logout': logout
        }

        with mock_results(methods):
            with mock.patch.dict(
                self.session.config,
                {'login_repeat': retry_amount}
            ):
                assert_raises(
                    webuntis.errors.NotLoggedInError, self.session._request,
                    'getCurrentSchoolyear')

        eq_(calls, expected_calls)

    def test_login_repeat_not_logged_in(self):
        retry_amount = 1
        calls = []

        # This produces a list of 5 * ['getCurrentSchoolyear'] with ['logout',
        # 'authenticate'] between each.
        expected_calls = ['authenticate', 'getCurrentSchoolyear']

        def authenticate(url, jsondata, headers):
            calls.append(jsondata['method'])
            return {
                'id': jsondata['id'],
                'result': {'sessionId': 'Foobar_session_' + jsondata['id']}
            }

        def _nope(url, jsondata, headers):
            calls.append(jsondata['method'])
            return {
                'id': jsondata['id'],
                'error': {'code': -8520, 'message': 'Not Logged In!'}
            }

        methods = {
            'authenticate': authenticate,
            'getCurrentSchoolyear': _nope,
            'logout': _nope
        }

        with mock_results(methods):
            with mock.patch.dict(
                self.session.config,
                {'login_repeat': retry_amount, 'jsessionid': None}
            ):
                assert_raises(
                    webuntis.errors.NotLoggedInError, self.session._request,
                    'getCurrentSchoolyear')

        eq_(calls, expected_calls)

    def test_session_contextmanager(self):
        s = webuntis.Session(**stub_session_parameters)

        with mock.patch('webuntis.Session.logout') as logout:
            with mock.patch('webuntis.Session.login', return_value=s) as login:
                assert webuntis.Session.login is login
                assert webuntis.Session.logout is logout

                with s.login():
                    pass

                login.assert_called_once_with()
                logout.assert_called_once_with(suppress_errors=True)

    def test_logout_not_logged_in(self):
        session_parameters = stub_session_parameters.copy()
        del session_parameters['jsessionid']

        s = webuntis.Session(**session_parameters)

        assert_raises(webuntis.errors.NotLoggedInError, s.logout)
        s.logout(suppress_errors=True)  # should not raise

        with mock.patch(
            'webuntis.Session._request'
        ):
            assert_raises(webuntis.errors.NotLoggedInError, s.logout)
            s.logout(suppress_errors=True)  # should not raise

    def test_login_no_creds(self):
        session_parameters = stub_session_parameters.copy()
        del session_parameters['jsessionid']
        del session_parameters['username']
        del session_parameters['password']

        s = webuntis.Session(**session_parameters)

        with mock.patch(
            'webuntis.Session._request',
            side_effect=Exception('Testing if login method raises AuthError '
                                  'due to invalid creds...')
        ):
            assert_raises(webuntis.errors.AuthError, s.login)

    def test_login_response_with_sessionid_and_without(self):
        session_parameters = stub_session_parameters.copy()
        del session_parameters['jsessionid']

        s = webuntis.Session(**session_parameters)

        def assert_call(request_mock):
            request_mock.assert_called_with(
                'authenticate',
                {
                    'user': session_parameters['username'],
                    'password': session_parameters['password'],
                    'client': session_parameters['useragent']
                },
                use_login_repeat=False
            )

        with mock.patch(
            'webuntis.Session._request',
            side_effect=[{'sessionId': '123456'}, {}]
        ) as request_mock:
            s.login()
            eq_(s.config['jsessionid'], '123456')
            eq_(request_mock.call_count, 1)
            assert_call(request_mock)

            del s.config['jsessionid']

            assert_raises(webuntis.errors.AuthError, s.login)
            assert 'jsessionid' not in s.config
            eq_(request_mock.call_count, 2)
            assert_call(request_mock)

    def test_custom_cachelen(self):
        cachelen = 13
        params = stub_session_parameters.copy()
        params['cachelen'] = cachelen
        s = webuntis.Session(**params)
        eq_(s.cache._maxlen, cachelen)
        assert 'cachelen' not in s.config


class InternalTests(OfflineTestCase):
    '''Test certain internal interfaces, such as utils'''

    def test_config_invalidattribute(self):
        assert 'nigglywiggly' not in self.session.config
        assert_raises(
            KeyError,
            self.session.config.__getitem__,
            'nigglywiggly'
        )

    def test_datetime_utils(self):
        obj = webuntis.utils.datetime_utils.parse_datetime(20121005, 0)
        eq_(obj.year, 2012)
        eq_(obj.month, 10)
        eq_(obj.day, 5)
        eq_(obj.hour, 0)
        eq_(obj.minute, 0)
        eq_(obj.second, 0)

    def test_filterdict_basic(self):
        store = webuntis.utils.FilterDict({
            'whatever': lambda x: x,
            'always_whoop': lambda x: 'whoop'
        })
        store['whatever'] = 'lel'
        eq_(store['whatever'], 'lel')

        store['always_whoop'] = 'what'
        eq_(store['always_whoop'], 'whoop')

    def test_filterdict_deletion_successful(self):
        store = webuntis.utils.FilterDict({
            'whatever': lambda x: x
        })

        store['whatever'] = 'HUE'

        del store['whatever']
        assert 'whatever' not in store
        assert_raises(KeyError, store.__getitem__, 'whatever')

    def test_filterdict_assigning_none_deletes(self):
        store = webuntis.utils.FilterDict({
            'whatever': lambda x: x
        })

        store['whatever'] = 'HUE'
        assert 'whatever' in store
        store['whatever'] = None
        assert 'whatever' not in store
        assert_raises(KeyError, store.__getitem__, 'whatever')

    def test_filterdict_getting_none_from_filter_deletes(self):
        return_val = True
        store = webuntis.utils.FilterDict({
            'whatever': lambda x: return_val
        })

        store['whatever'] = 'HUE'
        eq_(store['whatever'], True)
        assert 'whatever' in store

        return_val = None
        store['whatever'] = 'HUE'
        assert 'whatever' not in store
        assert_raises(KeyError, store.__getitem__, 'whatever')

    def test_session_invalidattribute(self):
        assert_raises(AttributeError, getattr, self.session, 'foobar')
        assert not hasattr(self.session, 'foobar')

    def test_requestcaching_no_io_on_second_time(self):
        jsonstr = get_json_resource('getklassen_mock.json')

        def getKlassen(url, jsondata, headers):
            return {'result': jsonstr}

        methods = {'getKlassen': getKlassen}

        with mock_results(methods):
            self.session.klassen(from_cache=True)

        self.session.klassen(from_cache=True)

        eq_(len(getKlassen.calls), 1)

    def test_requestcaching_no_caching_without_being_told_so(self):
        jsonstr = get_json_resource('getklassen_mock.json')

        def getKlassen(url, jsondata, headers):
            return {'result': jsonstr}

        class ThisVeryCustomException(Exception):
            pass

        def blow_up(*args, **kwargs):
            raise ThisVeryCustomException

        methods = {'getKlassen': getKlassen}

        failing_methods = {'getKlassen': blow_up}

        with mock_results(methods):
            self.session.klassen()
            self.session.klassen()

        eq_(len(getKlassen.calls), 2)

        with mock_results(failing_methods):
            assert_raises(ThisVeryCustomException, self.session.klassen)

    def test_no_cache_argument_in_caching_key(self):
        '''Any result wrapper method takes a ``cache`` arg which activates the
        cache if True. This test ensures that this arg doesn't show up in the
        cache dict key at the end.'''

        jsonstr = get_json_resource('getklassen_mock.json')

        def getKlassen(url, jsondata, headers):
            return {'result': jsonstr}

        with mock_results({'getKlassen': getKlassen}):
            self.session.klassen(from_cache='LELE')  # should be True-ish
            self.session.klassen(from_cache='YESPLS')  # should be too

        eq_(len(getKlassen.calls), 1)
        eq_(len(self.session.cache), 1)

    def test_listitem(self):
        session = self.session
        parent = None
        data = {'id': 42}
        item = webuntis.objects.ListItem(session=session, parent=parent,
                                         data=data)

        eq_(item._session, session)
        eq_(item._parent, parent)
        eq_(item._data, data)
        eq_(item.id, data['id'])
        eq_(int(item), item.id)

    def test_userinput_server(self):
        tests = [
            ('webuntis.grupet.at',
                'http://webuntis.grupet.at/WebUntis/jsonrpc.do'),
            ('https://webuntis.grupet.at',
                'https://webuntis.grupet.at/WebUntis/jsonrpc.do'),
            ('webuntis.grupet.at:8080',
                'http://webuntis.grupet.at:8080/WebUntis/jsonrpc.do'),
            ('webuntis.grupet.at/a/b/c', 'http://webuntis.grupet.at/a/b/c'),
            ('webuntis.grupet.at/', 'http://webuntis.grupet.at/'),
        ]

        for parser_input, expected_output in tests:
            eq_(
                webuntis.utils.userinput.server(parser_input),
                expected_output
            )

        assert_raises(ValueError, webuntis.utils.userinput.server, '!"$%')
        assert_raises(ValueError, webuntis.utils.userinput.server, '')

    def test_userinput_string(self):
        s = webuntis.utils.userinput.string
        try:
            string = unicode
        except NameError:  # pragma: no cover
            string = str

        assert type(s('foo')) is string

    def test_resultclass_invalid_arguments(self):
        assert_raises(TypeError, webuntis.objects.Result, session=self.session,
                      kwargs={}, data="LELELE")
        assert_raises(TypeError, webuntis.objects.Result)
        assert_raises(TypeError, webuntis.objects.Result, session=self.session)

    def test_jsonrpcrequest_parse_result_invalid_request_id_returned(self):
        method = 'getThingsNotExisting'
        params = {'are': 'you', 'seri': 'ous'}

        assert_raises(
            webuntis.errors.RemoteError,
            webuntis.utils.remote._parse_result,
            {
                'jsonrpc': '2.0',
                'method': method,
                'params': params,
                'id': '2012-03-14 13:37:12.345678'
            }, {
                'jsonrpc': '2.0',
                'id': 'COMPLETELYUNRELATED',
                'result': [1, 2, 3]
            }
        )

    def test_jsonrpcrequest_parse_error_code(self):
        method = 'getThingsNotExisting'
        params = {'are': 'you', 'seri': 'ous'}
        reqid = '2012-03-14 13:37:12.345678'

        def check():
            assert_raises(
                webuntis.errors.RemoteError,
                webuntis.utils.remote._parse_error_code,
                request_data, result_data
            )

        request_data = {
            'jsonrpc': '2.0',
            'id': reqid,
            'method': method,
            'params': params
        }

        result_data = {
            'jsonrpc': '2.0',
            'id': reqid,
            'error': {
                'code': 234567854323456789876543,
                'message': 'I am a teapot.'
            }
        }
        check()

        del result_data['error']
        check()

    def test_jsonrpcrequest_send_request_valid_request(self):
        self.request_patcher.stop()

        url = 'http://example.com'
        headers = {'User-agent': 'Netscape'}

        with mock_urlopen(b'{"ret": "VAL"}', expected_url=url,
                          expected_data=b'{"la": "LU"}',
                          expected_headers=headers):
            eq_(
                webuntis.utils.remote._send_request(
                    url,
                    {'la': 'LU'},
                    headers
                ),
                {'ret': 'VAL'}
            )

    def test_jsonrpcrequest_send_request_invalid_json_response(self):
        self.request_patcher.stop()

        url = 'http://example.com'
        headers = {'User-agent': 'Netscape'}

        with mock_urlopen(b'LOL DUDE', expected_url=url,
                          expected_data=b'{"la": "LU"}',
                          expected_headers=headers):
            assert_raises(
                webuntis.errors.RemoteError,
                webuntis.utils.remote._send_request,
                url,
                {'la': 'LU'},
                headers
            )

    def test_jsonrpcrequest_send_request_invalid_json_input(self):
        self.request_patcher.stop()

        url = 'http://example.com'
        headers = {'User-agent': 'Netscape'}

        with mock_urlopen(Exception('No.'), expected_url=url,
                          expected_data='Does not matter',
                          expected_headers=headers):
            assert_raises(
                TypeError,
                webuntis.utils.remote._send_request,
                url,
                object(),
                headers
            )

    def test_resultobject_invalid_params(self):
        valid_result = webuntis.objects.Result(data={}, parent=None,
                                               session=self.session)
        assert_raises(TypeError, webuntis.objects.Result, data={},
                      parent='WAT')
        assert_raises(TypeError, webuntis.objects.Result, data={},
                      parent=valid_result, session=self.session)

    def test_datetime_utils_date(self):
        dateint = 20121212
        datestr = str(dateint)
        dateobj = datetime.datetime.strptime(datestr, '%Y%m%d')
        eq_(webuntis.utils.datetime_utils.parse_date(dateint), dateobj)
        eq_(webuntis.utils.datetime_utils.parse_date(datestr), dateobj)
        eq_(webuntis.utils.datetime_utils.format_date(dateobj), dateint)

    def test_datetime_utils_time(self):
        timeint = 1337
        timestr = str(timeint)
        timeobj = datetime.datetime.strptime(timestr, '%H%M')

        eq_(webuntis.utils.datetime_utils.parse_time(timeint), timeobj)
        eq_(webuntis.utils.datetime_utils.parse_time(timestr), timeobj)
        eq_(webuntis.utils.datetime_utils.format_time(timeobj), timeint)

    def test_lrudict(self):
        d = webuntis.utils.LruDict(maxlen=3)
        d['foo'] = 2
        d['bar'] = 3
        d['baz'] = 4
        d['blah'] = 5
        eq_(len(d), 3)
        assert 'foo' not in d
        assert 'bar' in d
        assert 'baz' in d
        assert 'blah' in d

    def test_sessioncache_clear_by_method(self):
        d = webuntis.utils.SessionCache(maxlen=5)

        d[webuntis.utils.cache_key('timetable', {})] = 'POOP'
        d[webuntis.utils.cache_key('timetable', {'A': 'B'})] = 'POOP2'
        d[webuntis.utils.cache_key('klassen', {})] = 'POOP3'

        d.clear('timetable')
        eq_(len(d), 1)
        assert webuntis.utils.cache_key('klassen', {}) in d
        eq_(d[webuntis.utils.cache_key('klassen', {})], 'POOP3')

        d.clear()
        eq_(len(d), 0)

        # is that even used?
        eq_(type(self.session.cache), webuntis.utils.SessionCache)

    def test_lazyproperty_from_instance(self):
        meth_calls = []

        class FooBoo(object):
            @webuntis.utils.lazyproperty
            def some_method(self):
                meth_calls.append(True)
                return 42

        boo = FooBoo()
        eq_(boo.some_method, 42)
        assert boo.some_method is 42
        eq_(len(meth_calls), 1)

    def test_sessioncachekey_is_unique(self):
        Key = webuntis.utils.cache_key

        a = Key('klassen', {})
        a1 = Key('klassen', {})
        b = Key('klassen', {'g': 'G'})
        c = Key('teachers', {})

        eq_(a, a1)
        eq_(hash(a), hash(a1))

        assert_not_equal(a, b)
        assert_not_equal(hash(a), hash(b))

        assert_not_equal(a, c)
        assert_not_equal(hash(a), hash(c))

    def test_sessioncachekey_is_hashable(self):
        now = datetime.datetime.now()
        today = datetime.date.today()
        key = webuntis.utils.cache_key

        # The hash builtin will take care of us if the results aren't hashable.
        hash(key('getStuff', {'foo': 'bar'}))
        hash(key('getStuff', {}))
        hash(key('getStuff', None))
        hash(key('fooBar', {'start': now}))
        hash(key('fooBar', {'start': today}))

    def test_dont_cache_the_same_thing(self):
        jsonstr = get_json_resource('gettimetables_mock.json')
        today = datetime.datetime.now()
        today2 = datetime.datetime.now()

        assert_not_equal(today, today2)

        def getTimetable(url, jsondata, headers):
            return {'result': jsonstr}

        with mock_results({'getTimetable': getTimetable}):
            self.session.timetable(start=today, end=today, klasse=123,
                                   from_cache=True)
            eq_(len(getTimetable.calls), 1)
            self.session.timetable(start=today2, end=today2, klasse=123,
                                   from_cache=True)
            eq_(len(getTimetable.calls), 1)



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
import datetime


forms = {
    'date': ('%Y%m%d', 6),
    'time': ('%H%M', 4),
}


def parse_date(string):
    return _parse(string, 'date')


def parse_time(string):
    return _parse(string, 'time')


def parse_datetime(date, time):
    return datetime.datetime.combine(
        parse_date(date).date(),
        parse_time(time).time()
    )


def _parse(string, form):
    form_string, formlen = forms[form]
    return datetime.datetime.strptime(_satinize(string, form), form_string)


def format_date(obj):
    return _format(obj, 'date')


def format_time(obj):
    return _format(obj, 'time')


def _format(obj, form):
    form_string, formlen = forms[form]
    try:
        int(obj)  # is this already formatted?
    except TypeError:
        return int(datetime.datetime.strftime(obj, form_string))
    else:
        return int(obj)

def _satinize(raw_date, form):
    '''Convert a raw date/time integer or string to a string with fixed length.'''
    form_string, formlen = forms[form]
    raw_date = str(raw_date)

    while len(raw_date) < formlen:
        raw_date = '0' + raw_date

    return raw_date



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

try:
    from collections import OrderedDict  # Python >= 2.7
except ImportError:
    from ordereddict import OrderedDict  # from dependency "ordereddict"

try:
    import json  # Python >= 2.6
except ImportError:
    import simplejson as json  # from dependency "simplejson"

try:
    # Python 3
    import urllib.request as urlrequest
    import urllib.error as urlerrors  # pragma: no cover
    import urllib.parse as urlparse  # pragma: no cover
except ImportError:
    # Python 2
    import urllib2
    import urlparse
    urlrequest = urlerrors = urllib2



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals


class Error(Exception):
    '''Superclass for all `python-webuntis`-specific errors, never gets raised
    directly.'''


class RemoteError(Error, IOError):
    '''There was some kind of error while interacting with the server.'''
    pass


class MethodNotFoundError(RemoteError):
    '''The JSON-RPC method was not found. This really should not occur.'''
    pass


class AuthError(RemoteError):
    '''Errors while logging in.'''
    pass


class BadCredentialsError(AuthError, ValueError):
    '''Invalid or missing username or password.'''
    pass


class NotLoggedInError(AuthError):
    '''The session expired or we never logged in.'''



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals

import webuntis
import logging

from webuntis.tests.utils import TestCaseBase


class RemoteTests(TestCaseBase):
    '''
    DEPRECATED. This only should be used if you want to test the library
    actually against a public test-server, which will take a very long time and
    will stress that server unneccessarily.
    '''
    def setUp(self):
        self.session = webuntis.session.Session(
            school='demo_inf',
            server='webuntis.grupet.at:8080',
            username='api',
            password='api',
            useragent='WebUntis Unittests'
        ).login()
        print('\n' + '*' * 100)

    def tearDown(self):
        try:
            self.session.logout()
        except:
            logging.warning('Was not able to log out!')
        finally:
            self.session = None

    def test_timetableutils_table_emptyinput(self):
        self.assertEqual(len(webuntis.utils.timetable_utils.table([])), 0)

    ## OBJECTS

    def test_getdepartments(self):
        self.session.departments()

    def test_getholidays(self):
        self.session.holidays()

    def test_getklassen(self):
        print('TEST_GETKLASSEN')
        klassen = self.session.klassen()

        for klasse in klassen:
            self.assertEqual(klasse.__class__,
                             webuntis.objects.KlassenObject)

            klasse.id
            klasse.name
            klasse.long_name

    def test_getrooms(self):
        self.session.rooms()

    def test_getschoolyears(self):
        self.session.schoolyears()

    def test_getsubjects(self):
        self.session.subjects()

    def test_getteachers(self):
        print('TEST_GETTEACHERS')
        teachers = self.session.teachers()

        for teacher in teachers:
            self.assertEqual(teacher.__class__,
                             webuntis.objects.TeacherObject)

            teacher.fore_name
            teacher.name
            teacher.surname
            teacher.long_name
            teacher.id

    def test_gettimegrid(self):
        self.session.timegrid()

    def test_gettimetables(self):
        print('TEST_GETTIMETABLES')
        timetables = []
        for i, klasse in enumerate(self.session.klassen()):
            if i > 5:
                break
            timetable = self.session.timetable(klasse=klasse)
            print('Length of timetable: ' + str(len(timetable)))
            timetables.append(timetable)

    ## /OBJECTS

    def test_emptyfilter(self):
        print('TEST_EMPTYFILTER')
        for klasse, klasse_filtered, i in zip(
            self.session.klassen(),
            self.session.klassen().filter(),
            range(len(self.session.klassen()))
        ):
            self.assertEqual(klasse.id, klasse_filtered.id)

    def test_loggedout(self):
        self.session.logout()
        self.assertFalse('jsessionid' in self.session.config)
        self.assertRaises(webuntis.errors.AuthError, self.session.klassen)



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

# Test suites:
from webuntis.tests.offline import *
#from webuntis.tests.online import *



#!/usr/bin/env python
'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from setuptools import setup
from sys import version_info

dependencies = []

if version_info < (2,6):
    dependencies.append('simplejson')

if version_info < (2,7):
    dependencies.append('ordereddict')

setup(
    name='webuntis',
    version='0.1.7',
    author='Markus Unterwaditzer',
    author_email='markus@unterwaditzer.net',
    packages=['webuntis', 'webuntis.utils', 'webuntis.tests'],
    include_package_data=True,
    url='http://dev.unterwaditzer.net/python-webuntis/',
    license='new-style BSD',
    description='Bindings for WebUntis API',
    long_description=open('README.rst').read(),
    install_requires=dependencies,
    classifiers=[
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    test_suite='webuntis.tests'
)



'''
    This file is part of python-webuntis

    :copyright: (c) 2013 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
import re
from .logger import log
from .third_party import urlparse


def server(url):
    if not re.match(r'^http(s?)\:\/\/', url):  # if we just have the hostname
        log('debug', 'The URL given doesn\'t seem to be a valid URL, just '
                     'gonna prepend "http://"')

        # append the http prefix and hope for the best
        url = 'http://' + url

    urlobj = urlparse.urlparse(url)

    if not urlobj.scheme or not urlobj.netloc:
        # urlparse failed
        raise ValueError('Not a valid URL or hostname')

    if not re.match(r'^[a-zA-Z0-9\.\:-_]+$', urlobj.netloc):
        # That's not even a valid hostname
        raise ValueError('Not a valid hostname')

    if urlobj.path == '/':
        log('warning', 'You specified that the API endpoint should be "/".'
                       'That is uncommon. If you didn\'t mean to do so,'
                       'remove the slash at the end of your "server"'
                       'parameter.')

    return urlobj.scheme + \
        '://' + \
        urlobj.netloc + \
        (urlobj.path or '/WebUntis/jsonrpc.do')


def string(value):
    '''Make the string unicode'''
    return '' + value

config_keys = {
    'username': string,
    'password': string,
    'jsessionid': string,
    'school': string,
    'server': server,
    'useragent': string,
    'login_repeat': int
}



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
from webuntis import errors
from webuntis.utils import log
from webuntis.utils.third_party import json, urlrequest

import datetime


_errorcodes = {
    -32601: errors.MethodNotFoundError,
    -8504: errors.BadCredentialsError,
    -8520: errors.NotLoggedInError
}
'''The API-errorcodes python-webuntis is able to interpret, together with the
exception that will be thrown.'''


def rpc_request(config, method, params):
    '''
    A method for sending a JSON-RPC request.

    :param config: A dictionary containing ``useragent``, ``server``,
        ``school``, ``username`` and ``password``
    :type config: dict

    :param method: The JSON-RPC method to be executed
    :type method: str

    :param params: JSON-RPC parameters to the method (should be JSON
        serializable)
    :type params: dict
    '''

    url = config['server'] + \
        '?school=' + \
        config['school']

    headers = {
        'User-Agent': config['useragent'],
        'Content-Type': 'application/json'
    }

    request_body = {
        'id': str(datetime.datetime.today()),
        'method': method,
        'params': params,
        'jsonrpc': '2.0'
    }

    if method != 'authenticate':
        if 'jsessionid' not in config:
            raise errors.NotLoggedInError(
                'Don\'t have JSESSIONID. Did you already log out?')
        else:
            headers['Cookie'] = 'JSESSIONID=' + \
                config['jsessionid']

    log('debug', 'Making new request:')
    log('debug', 'URL: ' + url)
    log('debug', 'DATA: ' + str(request_body))

    result_body = _send_request(
        url,
        request_body,
        headers
    )
    return _parse_result(request_body, result_body)


def _parse_result(request_body, result_body):
    '''A subfunction of request, that, given the decoded JSON result,
    handles the error codes or, if everything went well, returns the result
    attribute of it. The request data has to be given too for logging and
    ID validation.

    :param request_body: The not-yet-encoded body of the request sent.
    :param result_body: The decoded body of the result recieved.
    '''

    if request_body['id'] != result_body['id']:
        raise errors.RemoteError(
            'Request ID was not the same one as returned.')

    try:
        return result_body['result']
    except KeyError:
        _parse_error_code(request_body, result_body)


def _parse_error_code(request_body, result_body):
    '''A helper function for handling JSON error codes.'''
    log('error', result_body)
    try:
        error = result_body['error']
        exc = _errorcodes[error['code']](error['message'])
    except KeyError:
        exc = errors.RemoteError(
            ('Some JSON-RPC-ish error happened. Please report this to the '
                'developer so he can implement a proper handling.'),
            str(result_body),
            str(request_body)
        )

    raise exc


def _send_request(url, data, headers):
    '''Sends a POST request given the endpoint URL, JSON-encodable data and
    a dictionary with headers.
    '''

    request = urlrequest.Request(url, json.dumps(data).encode(), headers)
    # this will eventually raise errors, e.g. if there's an unexpected http
    # status code
    result_obj = urlrequest.urlopen(request)
    result = result_obj.read().decode('utf-8')

    try:
        result_data = json.loads(result)
        log('debug', 'Valid JSON found')
    except ValueError:
        raise errors.RemoteError('Invalid JSON', str(result))
    else:
        return result_data
